@@ -1,7 +1,7 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+__all__ = ("mixin_class", "mixin_class_method")
 
-from __future__ import absolute_import
-
+import functools
 import sys
 
 import awkward as ak
@@ -51,7 +51,7 @@ def mixin_class(registry, name=None):
                     if rhs is None:
                         registry.setdefault((ufunc, behavior_name), method)
                         continue
-                    for rhs_name in list(rhs) + [behavior_name]:
+                    for rhs_name in [*list(rhs), behavior_name]:
                         registry.setdefault((ufunc, behavior_name, rhs_name), method)
                         if transpose is not None and rhs_name != behavior_name:
                             registry.setdefault(
@@ -64,14 +64,19 @@ def mixin_class(registry, name=None):
     return register
 
 
-def mixin_class_method(ufunc, rhs=None, transpose=True):
+def _call_transposed(func, left, right):
+    return func(right, left)
+
+
+def mixin_class_method(ufunc, rhs=None, *, transpose=True):
     """
     Args:
         ufunc (numpy.ufunc): A universal function (or NEP18 callable) that is
             hooked in Awkward Array, i.e. it can be the first argument of a behavior.
         rhs (Set[type] or None): Set of right-hand side argument types, optional
             if wrapping a unary function. The left-hand side is expected to
-            always be `self` of the parent class.
+            always be `self` of the parent class. The current class is implicitly
+            included in this set.
         transpose (bool): If true, automatically create a transpose signature
             (only makes sense for binary ufuncs).
 
@@ -83,28 +88,19 @@ def mixin_class_method(ufunc, rhs=None, transpose=True):
 
     def register(method):
         if not isinstance(rhs, (set, type(None))):
-            raise ValueError(
-                "expected a set of right-hand-side argument types"
-                + ak._util.exception_suffix(__file__)
+            raise ak._errors.wrap_error(
+                ValueError("expected a set of right-hand-side argument types")
             )
         if transpose and rhs is not None:
-
-            def transposed(left, right):
-                return method(right, left)
-
             # make a copy of rhs, we will edit it later
-            method._awkward_mixin = (ufunc, set(rhs), transposed)
+            # use partial & a module-scoped function so that this is pickleable
+            method._awkward_mixin = (
+                ufunc,
+                set(rhs),
+                functools.partial(_call_transposed, method),
+            )
         else:
             method._awkward_mixin = (ufunc, rhs, None)
         return method
 
     return register
-
-
-__all__ = [
-    x for x in list(globals()) if not x.startswith("_") and x not in ("sys", "ak")
-]
-
-
-def __dir__():
-    return __all__
