@@ -1,17 +1,19 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
-
-from __future__ import absolute_import
-
 import awkward as ak
+from awkward._behavior import behavior_of
+from awkward._layout import wrap_layout
+from awkward._nplikes import ufuncs
+from awkward._nplikes.numpylike import NumpyMetadata
+from awkward.highlevel import Array
 
-np = ak.nplike.NumpyMetadata.instance()
+np = NumpyMetadata.instance()
 
 
-class ByteBehavior(ak.highlevel.Array):
+class ByteBehavior(Array):
     __name__ = "Array"
 
     def __bytes__(self):
-        tmp = ak.nplike.of(self.layout).asarray(self.layout)
+        tmp = self.layout.backend.nplike.asarray(self.layout)
         if hasattr(tmp, "tobytes"):
             return tmp.tobytes()
         else:
@@ -24,8 +26,7 @@ class ByteBehavior(ak.highlevel.Array):
         return repr(self.__bytes__())
 
     def __iter__(self):
-        for x in self.__bytes__():
-            yield x
+        yield from self.__bytes__()
 
     def __eq__(self, other):
         if isinstance(other, (bytes, ByteBehavior)):
@@ -40,20 +41,24 @@ class ByteBehavior(ak.highlevel.Array):
         if isinstance(other, (bytes, ByteBehavior)):
             return bytes(self) + bytes(other)
         else:
-            raise TypeError("can only concatenate bytes to bytes")
+            raise ak._errors.wrap_error(
+                TypeError("can only concatenate bytes to bytes")
+            )
 
     def __radd__(self, other):
         if isinstance(other, (bytes, ByteBehavior)):
             return bytes(other) + bytes(self)
         else:
-            raise TypeError("can only concatenate bytes to bytes")
+            raise ak._errors.wrap_error(
+                TypeError("can only concatenate bytes to bytes")
+            )
 
 
-class CharBehavior(ak.highlevel.Array):
+class CharBehavior(Array):
     __name__ = "Array"
 
     def __bytes__(self):
-        tmp = ak.nplike.of(self.layout).asarray(self.layout)
+        tmp = self.layout.backend.nplike.asarray(self.layout)
         if hasattr(tmp, "tobytes"):
             return tmp.tobytes()
         else:
@@ -66,8 +71,7 @@ class CharBehavior(ak.highlevel.Array):
         return repr(self.__bytes__().decode("utf-8", "surrogateescape"))
 
     def __iter__(self):
-        for x in self.__str__():
-            yield x
+        yield from self.__str__()
 
     def __eq__(self, other):
         if isinstance(other, (str, CharBehavior)):
@@ -82,52 +86,47 @@ class CharBehavior(ak.highlevel.Array):
         if isinstance(other, (str, CharBehavior)):
             return str(self) + str(other)
         else:
-            raise TypeError("can only concatenate str to str")
+            raise ak._errors.wrap_error(TypeError("can only concatenate str to str"))
 
     def __radd__(self, other):
         if isinstance(other, (str, CharBehavior)):
             return str(other) + str(self)
         else:
-            raise TypeError("can only concatenate str to str")
+            raise ak._errors.wrap_error(TypeError("can only concatenate str to str"))
 
 
-ak.behavior["byte"] = ByteBehavior
-ak.behavior["__typestr__", "byte"] = "byte"
-ak.behavior["char"] = CharBehavior
-ak.behavior["__typestr__", "char"] = "char"
-
-
-class ByteStringBehavior(ak.highlevel.Array):
+class ByteStringBehavior(Array):
     __name__ = "Array"
 
     def __iter__(self):
-        for x in super(ByteStringBehavior, self).__iter__():
+        for x in super().__iter__():
             yield x.__bytes__()
 
 
-class StringBehavior(ak.highlevel.Array):
+class StringBehavior(Array):
     __name__ = "Array"
 
     def __iter__(self):
-        for x in super(StringBehavior, self).__iter__():
+        for x in super().__iter__():
             yield x.__str__()
 
 
-ak.behavior["bytestring"] = ByteStringBehavior
-ak.behavior["__typestr__", "bytestring"] = "bytes"
-ak.behavior["string"] = StringBehavior
-ak.behavior["__typestr__", "string"] = "string"
-
-
 def _string_equal(one, two):
-    nplike = ak.nplike.of(one, two)
-    behavior = ak._util.behaviorof(one, two)
+    behavior = behavior_of(one, two)
 
-    one, two = ak.without_parameters(one).layout, ak.without_parameters(two).layout
+    one, two = (
+        ak.operations.without_parameters(one).layout,
+        ak.operations.without_parameters(two).layout,
+    )
+    nplike = one.backend.nplike
 
     # first condition: string lengths must be the same
-    counts1 = nplike.asarray(one.count(axis=-1))
-    counts2 = nplike.asarray(two.count(axis=-1))
+    counts1 = nplike.asarray(
+        ak._do.reduce(one, ak._reducers.Count(), axis=-1, mask=False)
+    )
+    counts2 = nplike.asarray(
+        ak._do.reduce(two, ak._reducers.Count(), axis=-1, mask=False)
+    )
 
     out = counts1 == counts2
 
@@ -139,36 +138,31 @@ def _string_equal(one, two):
         onepossible = one[possible]
         twopossible = two[possible]
 
-        reduced = ak.all(ak.Array(onepossible) == ak.Array(twopossible), axis=-1).layout
-
+        reduced = ak.operations.all(
+            ak.Array(onepossible) == ak.Array(twopossible), axis=-1
+        ).layout
         # update same-length strings with a verdict about their characters
-        out[possible] = reduced
+        out[possible] = reduced.data
 
-    return ak._util.wrap(ak.layout.NumpyArray(out), behavior)
+    return wrap_layout(ak.contents.NumpyArray(out), behavior)
 
 
 def _string_notequal(one, two):
     return ~_string_equal(one, two)
 
 
-ak.behavior[ak.nplike.numpy.equal, "bytestring", "bytestring"] = _string_equal
-ak.behavior[ak.nplike.numpy.equal, "string", "string"] = _string_equal
-ak.behavior[ak.nplike.numpy.not_equal, "bytestring", "bytestring"] = _string_notequal
-ak.behavior[ak.nplike.numpy.not_equal, "string", "string"] = _string_notequal
-
-
 def _string_broadcast(layout, offsets):
-    nplike = ak.nplike.of(offsets)
-    offsets = nplike.asarray(offsets)
+    index_nplike = layout.backend.index_nplike
+    offsets = index_nplike.asarray(offsets)
     counts = offsets[1:] - offsets[:-1]
     if ak._util.win or ak._util.bits32:
-        counts = counts.astype(np.int32)
-    parents = nplike.repeat(nplike.arange(len(counts), dtype=counts.dtype), counts)
-    return ak.layout.IndexedArray64(ak.layout.Index64(parents), layout).project()
-
-
-ak.behavior["__broadcast__", "bytestring"] = _string_broadcast
-ak.behavior["__broadcast__", "string"] = _string_broadcast
+        counts = index_nplike.astype(counts, dtype=np.int32)
+    parents = index_nplike.repeat(
+        index_nplike.arange(counts.size, dtype=counts.dtype), counts
+    )
+    return ak.contents.IndexedArray(
+        ak.index.Index64(parents, nplike=index_nplike), layout
+    ).project()
 
 
 def _string_numba_typer(viewtype):
@@ -183,59 +177,57 @@ def _string_numba_typer(viewtype):
 def _string_numba_lower(
     context, builder, rettype, viewtype, viewval, viewproxy, attype, atval
 ):
+    import llvmlite.ir
     import numba
-    import llvmlite.llvmpy.core
 
-    whichpos = ak._connect._numba.layout.posat(
+    whichpos = ak._connect.numba.layout.posat(
         context, builder, viewproxy.pos, viewtype.type.CONTENT
     )
-    nextpos = ak._connect._numba.layout.getat(
+    nextpos = ak._connect.numba.layout.getat(
         context, builder, viewproxy.arrayptrs, whichpos
     )
 
-    whichnextpos = ak._connect._numba.layout.posat(
+    whichnextpos = ak._connect.numba.layout.posat(
         context, builder, nextpos, viewtype.type.contenttype.ARRAY
     )
 
-    startspos = ak._connect._numba.layout.posat(
+    startspos = ak._connect.numba.layout.posat(
         context, builder, viewproxy.pos, viewtype.type.STARTS
     )
-    startsptr = ak._connect._numba.layout.getat(
+    startsptr = ak._connect.numba.layout.getat(
         context, builder, viewproxy.arrayptrs, startspos
     )
     startsarraypos = builder.add(viewproxy.start, atval)
-    start = ak._connect._numba.layout.getat(
+    start = ak._connect.numba.layout.getat(
         context, builder, startsptr, startsarraypos, viewtype.type.indextype.dtype
     )
 
-    stopspos = ak._connect._numba.layout.posat(
+    stopspos = ak._connect.numba.layout.posat(
         context, builder, viewproxy.pos, viewtype.type.STOPS
     )
-    stopsptr = ak._connect._numba.layout.getat(
+    stopsptr = ak._connect.numba.layout.getat(
         context, builder, viewproxy.arrayptrs, stopspos
     )
     stopsarraypos = builder.add(viewproxy.start, atval)
-    stop = ak._connect._numba.layout.getat(
+    stop = ak._connect.numba.layout.getat(
         context, builder, stopsptr, stopsarraypos, viewtype.type.indextype.dtype
     )
 
-    baseptr = ak._connect._numba.layout.getat(
+    baseptr = ak._connect.numba.layout.getat(
         context, builder, viewproxy.arrayptrs, whichnextpos
     )
     rawptr = builder.add(
         baseptr,
-        ak._connect._numba.castint(
+        ak._connect.numba.layout.castint(
             context, builder, viewtype.type.indextype.dtype, numba.intp, start
         ),
     )
     rawptr_cast = builder.inttoptr(
         rawptr,
-        llvmlite.llvmpy.core.Type.pointer(
-            llvmlite.llvmpy.core.Type.int(numba.intp.bitwidth // 8)
-        ),
+        llvmlite.ir.PointerType(llvmlite.ir.IntType(numba.intp.bitwidth // 8)),
     )
     strsize = builder.sub(stop, start)
-    strsize_cast = ak._connect._numba.castint(
+    strsize_cast = ak._connect.numba.layout.castint(
         context, builder, viewtype.type.indextype.dtype, numba.intp, strsize
     )
 
@@ -257,16 +249,33 @@ def _string_numba_lower(
     return out
 
 
-ak.behavior["__numba_typer__", "bytestring"] = _string_numba_typer
-ak.behavior["__numba_lower__", "bytestring"] = _string_numba_lower
-ak.behavior["__numba_typer__", "string"] = _string_numba_typer
-ak.behavior["__numba_lower__", "string"] = _string_numba_lower
+def _cast_bytes_or_str_to_string(string):
+    return ak.to_layout([string])
 
 
-__all__ = [
-    x for x in list(globals()) if not x.startswith("_") and x not in ("ak", "np")
-]
+def register(behavior):
+    behavior["byte"] = ByteBehavior
+    behavior["__typestr__", "byte"] = "byte"
+    behavior["char"] = CharBehavior
+    behavior["__typestr__", "char"] = "char"
 
+    behavior["bytestring"] = ByteStringBehavior
+    behavior["__typestr__", "bytestring"] = "bytes"
+    behavior["string"] = StringBehavior
+    behavior["__typestr__", "string"] = "string"
 
-def __dir__():
-    return __all__
+    behavior[ufuncs.equal, "bytestring", "bytestring"] = _string_equal
+    behavior[ufuncs.equal, "string", "string"] = _string_equal
+    behavior[ufuncs.not_equal, "bytestring", "bytestring"] = _string_notequal
+    behavior[ufuncs.not_equal, "string", "string"] = _string_notequal
+
+    behavior["__broadcast__", "bytestring"] = _string_broadcast
+    behavior["__broadcast__", "string"] = _string_broadcast
+
+    behavior["__numba_typer__", "bytestring"] = _string_numba_typer
+    behavior["__numba_lower__", "bytestring"] = _string_numba_lower
+    behavior["__numba_typer__", "string"] = _string_numba_typer
+    behavior["__numba_lower__", "string"] = _string_numba_lower
+
+    behavior["__cast__", str] = _cast_bytes_or_str_to_string
+    behavior["__cast__", bytes] = _cast_bytes_or_str_to_string
